//
// Created by gwall on 1/10/18.
//

#include "video_encoder.h"
#include "parser_helpers.h"
#include "encoder_helpers.h"

using namespace std;

VideoEncoder::VideoEncoder(unsigned height, unsigned width, char *output_fname) {
    set_defaults(&(ctx));
    ctx.height = height;
    ctx.width = width;
    ctx.encoder_pixfmt = V4L2_PIX_FMT_H264;
    ctx.out_file_path = strdup(output_fname);
    setupContext();
    frame_count = 0;

}

VideoEncoder::~VideoEncoder() {
    cout << "VideoEncoder Self Destruct!!!" << endl;
    //abort(&ctx);
    ctx.enc->capture_plane.waitForDQThread(-1);
    cleanup(ctx, 1);

}

/*
a YUV buffer is 1.5x the size of a raw Y buffer
https://linuxtv.org/downloads/v4l-dvb-apis/uapi/v4l/pixfmt-yuv420m.html
*/
void VideoEncoder::compressYFrame(uint8_t *y_bytes) {
    if (y_bytes) {
        uint8_t *yuv_bytes = (uint8_t *) malloc((ctx.height * ctx.width * 3) / 2);
        //copy the raw luminance values into the new buffer
        memcpy(yuv_bytes, y_bytes, ctx.height * ctx.width);
        int offset = ctx.height * ctx.width;
        uint8_t blank_uv = 0x80;
        //fill in the U and V fields with blank chrominance values
        memset(yuv_bytes + offset, blank_uv, (ctx.height * ctx.width) / 2);
        compressYUV420Frame(yuv_bytes);
        free(yuv_bytes);
    } else {
        compressYUV420Frame(nullptr);
    }
}

/*
a YUV buffer is 1.5x the size of a raw Y buffer
https://linuxtv.org/downloads/v4l-dvb-apis/uapi/v4l/pixfmt-yuv420m.html
*/
void VideoEncoder::compressYUV420Frame(uint8_t *yuv_bytes) {
    encodeFrame(yuv_bytes);
}

int VideoEncoder::setupContext() {
    int ret = 0;

    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H265) {
        test_error(ctx.width < 144 || ctx.height < 144, "Height/Width should be"
                " > 144 for H.265", ctx);
    }

    if (ctx.use_gold_crc) {
        ctx.pBitStreamCrc = InitCrc(CRC32_POLYNOMIAL);
        test_error(!ctx.pBitStreamCrc, "InitCrc failed", ctx);
    }

    ctx.out_file = new ofstream(ctx.out_file_path);
    test_error(!ctx.out_file->is_open(), "Could not open output file", ctx);

    ctx.enc = NvVideoEncoder::createVideoEncoder("enc0");
    test_error(!ctx.enc, "Could not create encoder", ctx);

    // It is necessary that Capture Plane format be set before Output Plane
    // format.
    // Set encoder capture plane format. It is necessary to set width and
    // height on thr capture plane as well
    // https://devtalk.nvidia.com/default/topic/1027631/formatting-images-to-feed-into-nvvideoencoder-tegra-multimedia-api-/
    ret = ctx.enc->setCapturePlaneFormat(ctx.encoder_pixfmt, ctx.width,
                                         ctx.height, (ctx.width * ctx.height * 3) / 2);
    test_error(ret < 0, "Could not set output plane format", ctx);

    // Set encoder output plane format
    ret = ctx.enc->setOutputPlaneFormat(V4L2_PIX_FMT_YUV420M, ctx.width, ctx.height);
    test_error(ret < 0, "Could not set output plane format", ctx);

    ret = ctx.enc->setBitrate(ctx.bitrate);
    test_error(ret < 0, "Could not set encoder bitrate", ctx);

    ret = ctx.enc->setProfile(ctx.profile);
    test_error(ret < 0, "Could not set encoder profile", ctx);

    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H264) {
        ret = ctx.enc->setLevel(ctx.level);
        test_error(ret < 0, "Could not set encoder level", ctx);
    }

    ret = ctx.enc->setRateControlMode(ctx.ratecontrol);
    test_error(ret < 0, "Could not set encoder rate control mode", ctx);

    ret = ctx.enc->setIDRInterval(ctx.idr_interval);
    test_error(ret < 0, "Could not set encoder IDR interval", ctx);

    ret = ctx.enc->setIFrameInterval(ctx.iframe_interval);
    test_error(ret < 0, "Could not set encoder I-Frame interval", ctx);

    ret = ctx.enc->setFrameRate(ctx.fps_n, ctx.fps_d);
    test_error(ret < 0, "Could not set framerate", ctx);

    if (ctx.temporal_tradeoff_level) {
        ret = ctx.enc->setTemporalTradeoff(ctx.temporal_tradeoff_level);
        test_error(ret < 0, "Could not set temporal tradeoff level", ctx);
    }

    if (ctx.slice_length) {
        ret = ctx.enc->setSliceLength(ctx.slice_length_type,
                                      ctx.slice_length);
        test_error(ret < 0, "Could not set slice length params", ctx);
    }

    if (ctx.hw_preset_type) {
        ret = ctx.enc->setHWPresetType(ctx.hw_preset_type);
        test_error(ret < 0, "Could not set encoder HW Preset Type", ctx);
    }

    if (ctx.virtual_buffer_size) {
        ret = ctx.enc->setVirtualBufferSize(ctx.virtual_buffer_size);
        test_error(ret < 0, "Could not set virtual buffer size", ctx);
    }

    if (ctx.num_reference_frames) {
        ret = ctx.enc->setNumReferenceFrames(ctx.num_reference_frames);
        test_error(ret < 0, "Could not set num reference frames", ctx);
    }

    if (ctx.slice_intrarefresh_interval) {
        ret = ctx.enc->setSliceIntrarefresh(ctx.slice_intrarefresh_interval);
        test_error(ret < 0, "Could not set slice intrarefresh interval", ctx);
    }

    if (ctx.insert_sps_pps_at_idr) {
        ret = ctx.enc->setInsertSpsPpsAtIdrEnabled(true);
        test_error(ret < 0, "Could not set insertSPSPPSAtIDR", ctx);
    }

    if (ctx.num_b_frames != (uint32_t) -1) {
        ret = ctx.enc->setNumBFrames(ctx.num_b_frames);
        test_error(ret < 0, "Could not set number of B Frames", ctx);
    }

    if ((ctx.nMinQpI != (uint32_t) QP_RETAIN_VAL) ||
        (ctx.nMaxQpI != (uint32_t) QP_RETAIN_VAL) ||
        (ctx.nMinQpP != (uint32_t) QP_RETAIN_VAL) ||
        (ctx.nMaxQpP != (uint32_t) QP_RETAIN_VAL) ||
        (ctx.nMinQpB != (uint32_t) QP_RETAIN_VAL) ||
        (ctx.nMaxQpB != (uint32_t) QP_RETAIN_VAL)) {
        ret = ctx.enc->setQpRange(ctx.nMinQpI, ctx.nMaxQpI, ctx.nMinQpP,
                                  ctx.nMaxQpP, ctx.nMinQpB, ctx.nMaxQpB);
        test_error(ret < 0, "Could not set quantization parameters", ctx);
    }

    if (ctx.dump_mv) {
        ret = ctx.enc->enableMotionVectorReporting();
        test_error(ret < 0, "Could not enable motion vector reporting", ctx);
    }

    if (ctx.bnoIframe) {
        ctx.iframe_interval = ((1 << 31) + 1); /* TODO: how can we do this properly */
        ret = ctx.enc->setIFrameInterval(ctx.iframe_interval);
        test_error(ret < 0, "Could not set encoder I-Frame interval", ctx);
    }

    if (ctx.enableROI) {
        v4l2_enc_enable_roi_param VEnc_enable_ext_roi_ctrl;

        VEnc_enable_ext_roi_ctrl.bEnableROI = ctx.enableROI;
        ret = ctx.enc->enableROI(VEnc_enable_ext_roi_ctrl);
        test_error(ret < 0, "Could not enable ROI", ctx);
    }

    if (ctx.bReconCrc) {
        v4l2_enc_enable_reconcrc_param VEnc_enable_recon_crc_ctrl;

        VEnc_enable_recon_crc_ctrl.bEnableReconCRC = ctx.bReconCrc;
        ret = ctx.enc->enableReconCRC(VEnc_enable_recon_crc_ctrl);
        test_error(ret < 0, "Could not enable Recon CRC", ctx);
    }

    if (ctx.externalRPS) {
        v4l2_enc_enable_ext_rps_ctr VEnc_enable_ext_rps_ctrl;

        VEnc_enable_ext_rps_ctrl.bEnableExternalRPS = ctx.externalRPS;
        if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H264) {
            VEnc_enable_ext_rps_ctrl.bGapsInFrameNumAllowed = ctx.bGapsInFrameNumAllowed;
            VEnc_enable_ext_rps_ctrl.nH264FrameNumBits = ctx.nH264FrameNumBits;
        }
        if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H265) {
            VEnc_enable_ext_rps_ctrl.nH265PocLsbBits = ctx.nH265PocLsbBits;
        }
        ret = ctx.enc->enableExternalRPS(VEnc_enable_ext_rps_ctrl);
        test_error(ret < 0, "Could not enable external RPS", ctx);
    }

    if (ctx.externalRCHints) {
        v4l2_enc_enable_ext_rate_ctr VEnc_enable_ext_rate_ctrl;

        VEnc_enable_ext_rate_ctrl.bEnableExternalPictureRC = ctx.externalRCHints;
        VEnc_enable_ext_rate_ctrl.nsessionMaxQP = ctx.sMaxQp;

        ret = ctx.enc->enableExternalRC(VEnc_enable_ext_rate_ctrl);
        test_error(ret < 0, "Could not enable external RC", ctx);
    }

    // Query, Export and Map the output plane buffers so that we can read
    // raw data into the buffers
    ret = ctx.enc->output_plane.setupPlane(V4L2_MEMORY_MMAP, 10, true, false);
    test_error(ret < 0, "Could not setup output plane", ctx);

    // Query, Export and Map the output plane buffers so that we can write
    // encoded data from the buffers
    ret = ctx.enc->capture_plane.setupPlane(V4L2_MEMORY_MMAP, 10, true, false);
    test_error(ret < 0, "Could not setup capture plane", ctx);

    // output plane STREAMON
    ret = ctx.enc->output_plane.setStreamStatus(true);
    test_error(ret < 0, "Error in output plane streamon", ctx);

    // capture plane STREAMON
    ret = ctx.enc->capture_plane.setStreamStatus(true);
    test_error(ret < 0, "Error in capture plane streamon", ctx);

    ctx.enc->capture_plane.
            setDQThreadCallback(encoder_capture_plane_dq_callback);

    // startDQThread starts a thread internally which calls the
    // encoder_capture_plane_dq_callback whenever a buffer is dequeued
    // on the plane
    ctx.enc->capture_plane.startDQThread(&ctx);

    // Enqueue all the empty capture plane buffers
    for (uint32_t i = 0; i < ctx.enc->capture_plane.getNumBuffers(); i++) {
        struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];

        memset(&v4l2_buf, 0, sizeof(v4l2_buf));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        v4l2_buf.index = i;
        v4l2_buf.m.planes = planes;

        ret = ctx.enc->capture_plane.qBuffer(v4l2_buf, nullptr);
        if (ret < 0) {
            cerr << "Error while queueing buffer at capture plane" << endl;
            abort(&ctx);
            //goto cleanup;
            return cleanup(ctx, 1);
        }
    }

    return ret;
}

/*
inline void memset64(void *buffer, uint64_t value, size_t count) {
    const size_t m = count / 8;
    uint64_t *p = (uint64_t *) buffer;
    for (size_t i = 0; i < m; ++i, ++p)
        *p = value;
}
 */

void VideoEncoder::copyYuvToBuffer(uint8_t *yuv_bytes, NvBuffer &buffer) {

    uint32_t i, j, k;
    uint8_t *src = yuv_bytes;
    uint8_t *dst;


    for (i = 0; i < buffer.n_planes; i++) {

        NvBuffer::NvBufferPlane &plane = buffer.planes[i];
        uint32_t bytes_per_row = plane.fmt.bytesperpixel * plane.fmt.width;
        dst = plane.data;
        plane.bytesused = 0;

        for (j = 0; j < plane.fmt.height; j++) {
            for (k = 0; k < bytes_per_row; k++) {
                *dst = *src;
                dst++;
                src++;
            }
            dst += (plane.fmt.stride - bytes_per_row);
        }
        plane.bytesused = plane.fmt.stride * plane.fmt.height;
    }
}


/*
void VideoEncoder::copyYuvToBuffer(uint8_t *yuv_bytes, NvBuffer &buffer, bool blankUV) {

    uint32_t i, j, k;
    //uint8_t *src = yuv_bytes;
    //uint8_t *dst;

    uint64_t *src = (uint64_t*)yuv_bytes;
    uint64_t *dst;

    for (i = 0; i < buffer.n_planes; i++) {

            NvBuffer::NvBufferPlane &plane = buffer.planes[i];
            uint32_t bytes_per_row = plane.fmt.bytesperpixel * plane.fmt.width;
            //cout << i << " " << bytes_per_row << " " << plane.fmt.stride << endl;
            dst = (uint64_t *) plane.data;
            plane.bytesused = 0;

        if (i == 0) {
            for (j = 0; j < plane.fmt.height; j++) {
                for (k = 0; k < bytes_per_row / 8; k++) {
                    *dst = *src;
                    dst++;
                    src++;
                }
                dst += (plane.fmt.stride - bytes_per_row) / 8;
            }
        } else {
            if (blankUV) {
                const uint64_t blank_UV = 0x8080808080808080;
                memset64(dst, blank_UV, plane.fmt.stride * plane.fmt.height);
            } else {
                for (j = 0; j < plane.fmt.height; j++) {
                    for (k = 0; k < bytes_per_row / 8; k++) {
                        *dst = *src;
                        dst++;
                        src++;
                    }
                    dst += (plane.fmt.stride - bytes_per_row) / 8;
                }
            }
        }
        plane.bytesused = plane.fmt.stride * plane.fmt.height;
    }
}
 */

int VideoEncoder::encodeFrame(uint8_t *yuv_bytes) {
    int ret = 0;
    struct v4l2_buffer v4l2_buf;
    struct v4l2_plane planes[MAX_PLANES];
    NvBuffer *buffer = ctx.enc->output_plane.getNthBuffer(frame_count);

    memset(&v4l2_buf, 0, sizeof(v4l2_buf));
    memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

    v4l2_buf.m.planes = planes;

    if (frame_count < ctx.enc->output_plane.getNumBuffers()) {
        v4l2_buf.index = frame_count;

        if (yuv_bytes)
            copyYuvToBuffer(yuv_bytes, *buffer);
        else
            v4l2_buf.m.planes[0].bytesused = 0;

    } else {
        if (ctx.enc->output_plane.dqBuffer(v4l2_buf, &buffer, nullptr, 10) < 0) {
            cerr << "ERROR while DQing buffer at output plane" << endl;
            abort(&ctx);
            return cleanup(ctx, 1);
        }
    }

    if (ctx.runtime_params_str &&
        (ctx.enc->output_plane.getTotalQueuedBuffers() ==
         ctx.next_param_change_frame)) {
        set_runtime_params(&ctx);
        if (ctx.runtime_params_str)
            get_next_runtime_param_change_frame(&ctx);
    }

    if (frame_count >= ctx.enc->output_plane.getNumBuffers()) {
        if (yuv_bytes)
            copyYuvToBuffer(yuv_bytes, *buffer);
        else
            v4l2_buf.m.planes[0].bytesused = 0;
    }

    ret = ctx.enc->output_plane.qBuffer(v4l2_buf, nullptr);
    if (ret < 0) {
        cerr << "Error while queueing buffer at output plane" << endl;
        abort(&ctx);
        return cleanup(ctx, 1);
    }


    frame_count++;
    return ret;
}


